/*
Copyright (C) 2003-2007 Anders Ervik <calvin@countzero.no>
Copyright (C) 2006-2007 Leonard Ritter

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#pragma once

#include <vector>
#include <list>

namespace zzub {

struct mem_outstream : zzub::outstream {
	std::vector<char> &buffer;
	
	mem_outstream(std::vector<char> &b) : buffer(b) {}

	virtual void write(void *buffer, int size) {		
		char *charbuf = (char*)buffer;
		while (size--) {
			this->buffer.push_back(*charbuf++);
		}
	}
};

struct mem_instream : zzub::instream {
	long pos;
	std::vector<char> &buffer;
	
	mem_instream(std::vector<char> &b) : buffer(b) { pos = 0; }
	
	virtual void read(void *buffer, int size) {
		assert((pos + size) <= this->buffer.size());
		memcpy(buffer, &this->buffer[pos], size);
		pos += size;
	}

	virtual long position() {
		return pos;
	}
	
	virtual void seek(long offset, int origin) {
		if (origin == SEEK_SET) {
			assert(offset < this->buffer.size());
			pos = offset;
		} else if (origin == SEEK_CUR) {
			assert((pos + offset) <= this->buffer.size());
			pos += offset;
		} else if (origin == SEEK_END) {
			assert(((long)this->buffer.size() - offset) > 0);
			pos = this->buffer.size() - offset;
		} else {
			assert(0);
		}
	}

	virtual long size() { return buffer.size(); }
};

struct mem_archive : zzub::archive {
	typedef std::map<std::string, std::vector<char> > buffermap;
	typedef std::pair<std::string, std::vector<char> > bufferpair;
	buffermap buffers;
	
	std::list<mem_outstream *> outstreams;
	std::list<mem_instream *> instreams;
	
	~mem_archive() {
		for (std::list<mem_outstream *>::iterator i = outstreams.begin(); i != outstreams.end(); ++i) {
			delete *i;
		}
		for (std::list<mem_instream *>::iterator i = instreams.begin(); i != instreams.end(); ++i) {
			delete *i;
		}
	}
	
	std::vector<char> &get_buffer(const char *path) {
		buffermap::iterator i = buffers.find(path);
		if(i == buffers.end()) {
			buffers.insert(bufferpair(path, std::vector<char>()));
			i = buffers.find(path);
		}
		return i->second;
	}
	
	virtual outstream *get_outstream(const char *path) {
		buffers.insert(bufferpair(path, std::vector<char>()));
		buffermap::iterator i = buffers.find(path);
		assert(i != buffers.end());
		mem_outstream *mos = new mem_outstream(i->second);
		outstreams.push_back(mos);
		return mos;
	}
	virtual instream *get_instream(const char *path) {
		buffermap::iterator i = buffers.find(path);
		if (i == buffers.end())
			return 0;
		mem_instream *mis = new mem_instream(i->second);
		instreams.push_back(mis);
		return mis;
	}
};

} // namespace zzub

