/*
Copyright (C) 2003-2007 Anders Ervik <calvin@countzero.no>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <algorithm>
#include "common.h"
#include "tools.h"

namespace zzub {

envelope_entry::envelope_entry() {
	attack=0;
	decay=0;
	sustain=0x4000;
	release=0x1000;
	subDivide=10;
	flags=21;
	disabled=true;

	points.push_back(envelope_point());
	points.back().x=0;
	points.back().y=65535;
	points.back().flags=0;

	points.push_back(envelope_point());
	points.back().x=65535;
	points.back().y=0;
	points.back().flags=0;
}


void envelope_entry::clear() {
	points.clear();
}


/***

	wave_level_ex

***/

wave_level_ex::wave_level_ex() {
    this->level=0;
	this->samples=0;
	this->sample_count=0;
    this->loop_start=0;
    this->loop_end=0;
    this->samples_per_second=0;
    this->wave=0;
}


/***

	wave_info_ex

***/

wave_info_ex::wave_info_ex() {
	flags=0;
	volume=0;
	envelopes.push_back(envelope_entry());
}

wave_info_ex::~wave_info_ex() {
	clear();
}

bool wave_info_ex::saveLevel(size_t level, const char* fn) {
	/*int f=::open(fn, _O_BINARY|_O_RDWR|_O_CREAT, _S_IREAD | _S_IWRITE );
	if (f==-1) return false;
	CWaveLevel& wl=levels[level];
	::write(f, wl.pSamples, wl.numSamples*sizeof(short)*(Flags&WF_STEREO?2:1));
	::close(f);
	return true;*/
	return false;
}

void wave_info_ex::clear() {
	this->volume=0;
	this->flags=0;
	this->fileName="";
	this->name="";
	for (size_t i=0; i<levels.size(); i++) {
		if (levels[i].samples!=0)
			delete[] levels[i].samples;	// wtf! this was delete for a long while before delete[]!!
		levels[i].sample_count=0;
	}
	levels.clear();
	envelopes.clear(); // if we do this...
	envelopes.push_back(envelope_entry()); // ...we also need to do this
}

bool wave_info_ex::isStereo() const {
	return flags&zzub::wave_flag_stereo?true:false;
}

void wave_info_ex::setStereo(bool state) {
	DWORD f=((DWORD)flags)&(0xFFFFFFFF^zzub::wave_flag_stereo);

	if (state)
		flags=f|zzub::wave_flag_stereo; else
		flags=f;
}

bool wave_info_ex::isExtended() const {
	return flags&zzub::wave_flag_extended?true:false;
}

size_t wave_info_ex::getLevels() const {
	return levels.size();
}

zzub::wave_level_ex* wave_info_ex::getLevel(size_t level) const {
	if (level<0 || level>=levels.size()) return 0;
	return &(zzub::wave_level_ex&)levels[level];
}

void wave_info_ex::setExtendedMode() {	// reallocated buffer for extension
	if (isExtended()) return ;
	
	// TODO: this only clears 8 first shorts, should reallocate sample with + 8 shorts
	flags|=zzub::wave_flag_extended;
	for (size_t i=0; i<levels.size(); i++) {
		if (levels[i].sample_count<8) continue;
		memset(levels[i].samples, 0, 8);
	}
}


// reallocate reallocates and truncates or inserts silence
bool wave_info_ex::reallocateLevel(size_t level, size_t samples) {

	size_t samplesIn16Bit=0;
	size_t waveChannels=this->isStereo()?2:1;
	size_t bytesPerSample=this->getBitsPerSample(level)/8;
	size_t oldSamples=getNumSamples(level);
	size_t oldSamplesIn16Bit=0;

	if (isExtended()) {
		// loopstart should be in 16-bit values
		samplesIn16Bit=((samples * bytesPerSample) / 2) + (4/waveChannels);
		oldSamplesIn16Bit=((oldSamples * bytesPerSample) / 2) + (4/waveChannels);
	} else {
		samplesIn16Bit=samples;
		oldSamplesIn16Bit=oldSamples;
	}


	
	zzub::wave_level_ex* l=&this->levels[level];
	short* pSamples=(short*)new short[samplesIn16Bit*waveChannels];

	if (l->samples) {
		size_t copySize=std::min(oldSamplesIn16Bit, samplesIn16Bit)*waveChannels;
		memcpy(pSamples, l->samples, copySize*sizeof(short));
		delete[] l->samples;
	} else {
		memset(pSamples, 0, sizeof(samplesIn16Bit));
	}

	l->sample_count=samplesIn16Bit;//numSamples;
	l->samples=pSamples;//(short*)new char[waveBufferSize];

	return true;
}


// bitsperSample = 16, 24 or 32 (currently only 16 supported )
// 8 bits are not supported internally and must be converted to 16 bit
bool wave_info_ex::allocateLevel(size_t level, size_t numSamples, int waveFormat) {
	if (level<0 || level>=0xc8) return false;

	if (level>=levels.size()) {
		levels.resize(level+1);
		for (size_t i=0; i < levels.size(); ++i)
		{
			levels[i].wave = this; // make sure they all carry a pointer back
			levels[i].level = i; // and an index
		}
	}

	int waveChannels=isStereo()?2:1;

	size_t waveBufferSize=0;
	size_t samplesIn16bit=0;

	bool allocExtended=false;

	switch (waveFormat)  {
		case zzub::wave_buffer_type_si16:
			waveBufferSize=numSamples*waveChannels*2;
			samplesIn16bit=numSamples;
			break;
		case zzub::wave_buffer_type_si24:
			waveBufferSize=numSamples*waveChannels*3;
			samplesIn16bit=(numSamples*3) / 2;
			allocExtended=true;
			break;
		case zzub::wave_buffer_type_f32:
			waveBufferSize=numSamples*waveChannels*sizeof(float);
			samplesIn16bit=numSamples*2;
			allocExtended=true;
			break;
		case zzub::wave_buffer_type_si32:
			waveBufferSize=numSamples*waveChannels*sizeof(DWORD);
			samplesIn16bit=numSamples*2;
			allocExtended=true;
			break;
		default:
			// unsupported number of bits
			return false;
	}
	if (allocExtended) {
		waveBufferSize+=8;	// 8 byte offset
		samplesIn16bit+=4/waveChannels; // and 8 bytes of sample data
	}

	zzub::wave_level_ex* l=&levels[level];
	if (l->samples)
		delete[] l->samples;

	l->sample_count=samplesIn16bit;//numSamples;
	l->samples=(short*)new char[waveBufferSize];
	l->samples_per_second=44100;
	l->root_note=note_value_c4;
	memset(l->samples, 0, waveBufferSize);

	if (allocExtended) {
		flags|=zzub::wave_flag_extended;	// ensure extended flag is set
		l->samples[0]=waveFormat;//bitsPerSample/8;
	}

	return true;
}

void* wave_info_ex::getSampleData(size_t level) const {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return 0;
	
	// TODO: should probably return something else to machines via MICallbacks (which it does?? because it reads directly from the level )
	if (isExtended()) {
		int waveChannels=isStereo()?2:1;
		// AHH!! eller .. nei?? er det forskjellig offset avhengig av antall kanaler?
		return &l->samples[4];//[4/waveChannels];	// +4 = +4 shorts = 8 bytes
	} else
		return l->samples;
}


size_t wave_info_ex::getBitsPerSample(size_t level) const {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return 0;

	if (isExtended()) {
		switch (l->samples[0]) {
			case zzub::wave_buffer_type_si16:
				return 16;
			case zzub::wave_buffer_type_si24:
				return 24;
			case zzub::wave_buffer_type_f32:
			case zzub::wave_buffer_type_si32:
				return 32;
			default:
#if defined(_WIN32)
				MessageBox(0, "Unknown sample format, press OK to debug", 0, MB_OK);			
				DebugBreak();
#else
				printf("Unknown sample format: %i\n", l->samples[0]);
#endif
				return 16;
		}
	} else
		return 16;
}

// spm er om det skal vre -4/channels, eller bare -4
size_t extendedSamples(size_t bytesPerSample, size_t wtSamples, size_t channels) {
	return ((wtSamples-(4/channels)) *2 ) / bytesPerSample;
}

int wave_info_ex::getWaveFormat(size_t level) const {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return 0;
	if (isExtended()) {
		return l->samples[0];
	} else
		return zzub::wave_buffer_type_si16;
}

size_t wave_info_ex::getNumSamples(size_t level) const {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return 0;
	if (isExtended()) {
		return extendedSamples(getBitsPerSample(level)/8, l->sample_count, isStereo()?2:1);
	} else
		return l->sample_count;
}

int wave_info_ex::getRootNote(size_t level) const {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return 0;
	return l->root_note;
}

size_t wave_info_ex::getLoopStart(size_t level) const {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return 0;
	if (isExtended())
		return extendedSamples(getBitsPerSample(level)/8, l->loop_start, isStereo()?2:1); else
		return l->loop_start;
}

size_t wave_info_ex::getLoopEnd(size_t level) const {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return 0;
	if (isExtended()) 
		return extendedSamples(getBitsPerSample(level)/8, l->loop_end, isStereo()?2:1); else
		return l->loop_end;
}

size_t wave_info_ex::getSamplesPerSec(size_t level) const {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return 0;
	return l->samples_per_second;
}

void wave_info_ex::setRootNote(size_t level, size_t value) {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return ;
	levels[level].root_note=value;
}

void wave_info_ex::setLoopStart(size_t level, size_t value) {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return ;
	if (isExtended()) {
		// loopstart should be in 16-bit values
		size_t bytesPerSample=this->getBitsPerSample(level)/8;
		size_t waveChannels=this->isStereo()?2:1;
		levels[level].loop_start=((value * bytesPerSample) / 2) + (4/waveChannels);
	} else {
	// TODO: handle extended waves
		levels[level].loop_start=value;
	}
}

void wave_info_ex::setLoopEnd(size_t level, size_t value) {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return ;
	// TODO: handle extended waves
	if (isExtended()) {
		// loopstart should be in 16-bit values
		size_t bytesPerSample=this->getBitsPerSample(level)/8;
		size_t waveChannels=this->isStereo()?2:1;
		levels[level].loop_end=((value * bytesPerSample) / 2) + (4/waveChannels);
	} else {
	// TODO: handle extended waves
		levels[level].loop_end=value;
	}
	//levels[level].LoopEnd=value;
}

void wave_info_ex::setSamplesPerSec(size_t level, size_t value) {
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return ;
	levels[level].samples_per_second=value;
}

// these methods are for raw wave management, insertWaveAt promises conversion 
// according to its waveFormat parameter, which makes me want to consider having a target
// waveFormat parameter for createWaveRange as well

bool wave_info_ex::createWaveRange(size_t level, size_t fromSample, size_t numSamples, void** sampleData) {
	if (fromSample<0 || !numSamples) return false;
	zzub::wave_level_ex* l=getLevel(level);
	if (!l) return false;

	size_t bytesPerSample=getBitsPerSample(level)/8;
	size_t channels=isStereo()?2:1;
	size_t size=numSamples*bytesPerSample*channels;

	*sampleData=new char[size];

	// find offset in source buffer from where we start copying
	size_t ofs=fromSample*bytesPerSample*channels;
	if (isExtended())
		ofs+=8;
	char* cbuffer=(char*)l->samples;
	void* srcbuf=&cbuffer[ofs];
	memcpy(*sampleData, srcbuf, size);

	return true;
}

bool wave_info_ex::silenceWaveRange(size_t level, size_t fromSample, size_t numSamples) {
	return true;
}

bool wave_info_ex::removeWaveRange(size_t level, size_t fromSample, size_t numSamples) {
	size_t size=this->getNumSamples(level);

	void* range2;
	size_t ch2=isStereo()?2:1;
    int format2=getWaveFormat(level);

	size_t range2Len=size - (fromSample+numSamples);
	createWaveRange(level, fromSample+numSamples, range2Len, &range2);

	size_t newSize=size-numSamples;
	reallocateLevel(level, newSize);

	insertWaveAt(level, fromSample, range2, ch2, format2, range2Len);

	// adjust end looping point
	size_t loopEnd=getLoopEnd(level);
	if (loopEnd>newSize)
		setLoopEnd(level, newSize);

	return true;
}


size_t waveFormatToBitSize(int waveFormat) {
	switch (waveFormat) {
		case zzub::wave_buffer_type_si16:
			return 16;
		case zzub::wave_buffer_type_si24:
			return 24;
		case zzub::wave_buffer_type_f32:
		case zzub::wave_buffer_type_si32:
			return 32;
		default:
			return 16;
	}
}

// TODO: make so we can paste in one channel!

bool wave_info_ex::insertWaveAt(size_t level, size_t atSample, void* sampleData, size_t channels, int fromFormat, size_t numSamples) {
	size_t waveChannels=isStereo()?2:1;

	zzub::wave_level_ex* l=&levels[level];

	void* tempbuffer=0;
	void* srcbuffer=0;

	size_t srcBytesPerSample=(waveFormatToBitSize(fromFormat)/8)*2;

	if (channels==1) {
		if (waveChannels==1) {
			// same number of channels
			srcbuffer=sampleData;
		} else
		if (waveChannels==2) {
			// need to convert from stereo to mono while pasting
			// allocate a temp stereo buffer, and convert the input wave to this
			srcbuffer=tempbuffer=new char[srcBytesPerSample*numSamples*2];
			CopyMonoToStereoEx(srcbuffer, sampleData, numSamples, fromFormat);
		}
	} else 
	if (channels==2) {

		if (waveChannels==1) {
			// need to convert from mono to stereo while pasting
			// allocate a temp mono buffer, and convert the input wave to this
			srcbuffer=tempbuffer=new char[srcBytesPerSample*numSamples];
			CopyStereoToMonoEx(srcbuffer, sampleData, numSamples, fromFormat);
		} else
		if (waveChannels==2) {
			srcbuffer=sampleData;
			// same number of channels
		}
	} else
		return false;

	char* cbuffer=(char*)l->samples;

	size_t bytesPerSample=(this->getBitsPerSample(level)/8) * waveChannels;

	size_t ofs=atSample*bytesPerSample;
	if (isExtended())
		ofs+=8;
	void* targetbuffer=&cbuffer[ofs];
	// we have a buffer which has the same number of channels as we need, now convert

	int waveFormat=getWaveFormat(level);
	switch (fromFormat) {
		case zzub::wave_buffer_type_si16:
			switch (waveFormat) {
				case zzub::wave_buffer_type_si16:
					// 16 to 16
					memcpy(targetbuffer, srcbuffer, numSamples*waveChannels*sizeof(short));
					break;
				case zzub::wave_buffer_type_si24:
					// 16 to 24
					Copy16To24(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
				case zzub::wave_buffer_type_f32:
					// 16 to 32 bit floating point
					Copy16ToF32(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
				case zzub::wave_buffer_type_si32:
					// 16 to 32 bit integer
					Copy16ToS32(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
			}
			break;
		case zzub::wave_buffer_type_si24:
			switch (waveFormat) {
				case zzub::wave_buffer_type_si16:
					// 24 to 16
					Copy24To16(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
				case zzub::wave_buffer_type_si24:
					// 24 to 24
					memcpy(targetbuffer, srcbuffer, numSamples*waveChannels*3);
					break;
				case zzub::wave_buffer_type_f32:
					// 24 to 32 bit floating point
					Copy24ToF32(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
				case zzub::wave_buffer_type_si32:
					// 24 to 32 bit integer
					Copy24ToS32(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
			}
			break;
		case zzub::wave_buffer_type_si32:
			switch (waveFormat) {
				case zzub::wave_buffer_type_si16:
					// integer 32 to 16
					CopyS32To16(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
				case zzub::wave_buffer_type_si24:
					// integer 32 to 24
					CopyS32To24(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
				case zzub::wave_buffer_type_f32:
					// integer 32 to 32 bit floating point
					CopyS32ToF32(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
				case zzub::wave_buffer_type_si32:
					// integer 32 to 32 bit integer
					memcpy(targetbuffer, srcbuffer, numSamples*waveChannels*4);
					break;
			}
			break;
		case zzub::wave_buffer_type_f32:
			switch (waveFormat) {
				case zzub::wave_buffer_type_si16:
					// float 32 to 16
					CopyF32To16(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
				case zzub::wave_buffer_type_si24:
					// float 32 to 24
					CopyF32To24(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
				case zzub::wave_buffer_type_f32:
					// float 32 to 32 bit floating point
					memcpy(targetbuffer, srcbuffer, numSamples*waveChannels*4);
					break;
				case zzub::wave_buffer_type_si32:
					// float 32 to 32 bit integer
					CopyF32ToS32(targetbuffer, srcbuffer, numSamples*waveChannels);
					break;
			}
			break;
	}

	// if we allocated a temp buffer, release temporary memory:
	if (tempbuffer)
		delete[] (char*)tempbuffer;

	return true;
}

size_t wave_info_ex::getLevelIndex(zzub::wave_level_ex* level) {
	for (size_t i=0; i<getLevels(); i++) {
		if (&levels[i]==level) return i;
	}
	return -1;
}


void wave_info_ex::setLooping(bool state) {
	if (state) {
		flags |= zzub::wave_flag_loop;
	} else {
		flags ^= flags & zzub::wave_flag_loop;
	}
}

void wave_info_ex::setBidir(bool state) {
	if (state) {
		flags |= zzub::wave_flag_pingpong;
	} else {
		flags ^= flags & zzub::wave_flag_pingpong;
	}
}


bool wave_info_ex::isLooping() {
	return flags&zzub::wave_flag_loop?true:false;
}

bool wave_info_ex::isBidir() {
	return flags&zzub::wave_flag_pingpong?true:false;
}



/***

	WaveTable

***/

wave_table::wave_table(void) {
	waves.resize(0xc8);	// max 200 waves
}

wave_table::~wave_table(void) {
	clear();
	waves.clear();
}

void wave_table::clear() {
	for (size_t i=0; i<waves.size(); i++) {
		waves[i].clear();
	}
}


};
